import os
from typing import Type, List

from django import template
from django.conf import settings
from django.db.models import QuerySet
from django.db.models.fields.files import FieldFile
from django.db.models.functions import Lower
from django.forms.boundfield import BoundField
from django.forms.fields import FileField
from django.template import loader
from django.templatetags.static import static as dstatic
from django.utils.html import escape
from django.utils.safestring import SafeText
from pursuitlib import utils
from pursuitlib.color import is_dark
from pursuitlib.utils import is_null_or_empty

from pursuitlib_django.forms import Form, DEFAULT_FORM_SECTION
from pursuitlib_django import utils

# Read the Django configuration

VERSION = getattr(settings, "VERSION")
DISPLAY_VERSION = getattr(settings, "DISPLAY_VERSION", VERSION)
BASE_URL = getattr(settings, "BASE_URL")
MEDIA_URL = getattr(settings, "MEDIA_URL")


register = template.Library()


@register.filter
def at(collection, key):
    if key in collection:
        return collection[key]
    elif (isinstance(collection, tuple) or isinstance(collection, list)) and key is int and 0 <= key < len(collection):
        return collection[key]
    return collection[key]


@register.filter
def contains(collection, value) -> bool:
    return value in collection


@register.filter
def concat(value1, value2) -> str:
    return str(value1) + str(value2)


@register.filter
def orderby(query: QuerySet, order_by: str):
    return query.order_by(order_by)


@register.filter
def iorderby(query: QuerySet, order_by: str):
    return query.order_by(Lower(order_by))


@register.simple_tag
def percentage(value: int, total: int, rounding: bool = False):
    if total == 0:
        return "0%"
    elif rounding:
        return str((value * 100 // total)) + "%"
    else: return str((value * 100 / total)) + "%"


@register.filter
def bool_text(value) -> str:
    if value is None:
        return "Inconnu"
    elif value:
        return "Oui"
    else: return "Non"


@register.filter
def as_widget(form_field: BoundField) -> SafeText:
    if isinstance(form_field.field, FileField):
        html = str(form_field.as_widget())
        offset = html.index('<input type="file"')
        has_checkbox = "checkbox" in html

        main = html[offset:]
        checkbox = '</div><div class="form-check m-2">\
                            <input type="checkbox" class="form-check-input" name="' + form_field.name + '-clear" id="' + form_field.name + '-clear_id">\
                            <label class="form-check-label" for="' + form_field.name + '-clear_id">Supprimer</label>'

        return SafeText(main + (checkbox if has_checkbox else ''))
    elif form_field.widget_type == "checkbox":
        return SafeText('<div class="form-check m-2">' + form_field.as_widget() +
                        '<label class="form-check-label" for="' + form_field.id_for_label + '">' + form_field.label + '</label></div>')
    else: return form_field.as_widget()


@register.filter
def unsafe_text(text: str) -> SafeText:
    return SafeText(escape(text).replace("\n", "<br>"))


@register.simple_tag
def static(path: str) -> str:
    return dstatic(path) + "?v=" + VERSION  # Invalidate the client static files cache after every update


@register.filter
def media_url(file: FieldFile) -> str:
    return MEDIA_URL + file.name


@register.filter
def full_url(url: str) -> str:
    return BASE_URL + url


@register.simple_tag
def version() -> str:
    return DISPLAY_VERSION


@register.simple_tag
def versnum() -> str:
    return VERSION


@register.simple_tag
def env(varname: str) -> str:
    return os.getenv(varname)


@register.filter
def alert_type(severity: str) -> str:
    if severity == "debug":
        return "dark"
    elif severity == "info":
        return "info"
    elif severity == "success":
        return "success"
    elif severity == "warning":
        return "warning"
    elif severity == "error":
        return "danger"
    else: return "light"


@register.filter
def not_empty(value) -> bool:
    return not is_null_or_empty(value)


@register.filter
def empty_none(value) -> str:
    if value is None:
        return ""
    else: return value


@register.filter
def filename(path: str) -> str:
    return os.path.basename(path)


@register.filter
def typeof(obj) -> Type:
    return type(obj)


@register.filter
def verbose_name(obj) -> str:
    return utils.get_verbose_name(obj)


@register.filter
def verbose_name_plural(obj) -> str:
    return utils.get_verbose_name_plural(obj)


@register.filter
def is_feminine(obj) -> bool:
    return utils.is_feminine(obj)


@register.filter
def genderize(obj, suffix: str = "e") -> str:
    return suffix if is_feminine(obj) else ""


@register.filter
def tostring(obj) -> str:
    return str(obj)


@register.filter()
def liststring(objects) -> str:
    return utils.list_to_string(objects)


@register.filter
def foreground(color: str) -> str:
    return "#FFFFFF" if is_dark(color) else "#000000"


@register.filter
def section_fields(form: Form, section: str) -> List[BoundField]:
    return form.get_section(section)

@register.simple_tag(takes_context=True)
def global_csrf_token(context) -> SafeText:
    csrf_token = context.get("csrf_token")
    if not csrf_token or csrf_token == "NOTPROVIDED":
        return SafeText("")
    else:
        return SafeText(f'<meta data-post data-post-csrfmiddlewaretoken="{csrf_token}">')


# Components

@register.simple_tag
def form_section(form: Form, section: str = DEFAULT_FORM_SECTION, padding: bool = True, label: bool = True):
    return loader.get_template("pursuitlib/component/form-section.html").render({"form": form,
                                                                                 "section": section,
                                                                                 "padding": padding,
                                                                                 "display_label": label})


@register.simple_tag
def field(form_field: BoundField, padding: bool = True, label: bool = True):
    return loader.get_template("pursuitlib/component/form-field.html").render({"field": form_field,
                                                                               "padding": padding,
                                                                               "display_label": label})


@register.simple_tag(takes_context=True)
def pagination(context, pages: dict = None, page_var: str = "page"):
    if pages is None:
        # Using the "pages" context variable by default
        pages = context["pages"]
    return loader.get_template("pursuitlib/component/pagination.html").render({"pages": pages,
                                                                               "page_var": page_var})



from typing import Type

from django.http import FileResponse
from django.utils.http import content_disposition_header

from pursuitlib_django.forms import Form


def view_form(form_type: Type[Form]):
    return lambda request, **kwargs: form_type.handle_request(request, **kwargs)


def serve_text(text: str, content_type: str = "text/plain", name: str = None) -> FileResponse:
    # When a text content type is used, the text needs to be passed as is.
    # It will then be UTF8-encoded
    return serve_file(text, f"{content_type}; charset=utf-8", name)


def serve_file(content, content_type: str = "application/octet-stream", name: str = None) -> FileResponse:
    resp = FileResponse(content, content_type=content_type, as_attachment=(name is not None), filename=name)
    # Django bugfix: set the Content-Disposition header manually
    resp.headers["Content-Disposition"] = content_disposition_header(resp.as_attachment, resp.filename)
    return resp



from typing import List, Optional, Dict, Type

import django.forms as dforms
from django.conf import settings
from django.contrib import messages
from django.db.models import Model
from django.forms import BoundField
from django.forms.utils import ErrorDict, ErrorList
from django.forms.widgets import CheckboxInput, TextInput, Select
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import render
from django.utils.safestring import SafeText
from django_select2.forms import Select2Mixin
from pursuitlib.utils import is_null_or_empty, get_oneline_string
from pursuitlib_django.utils import get_verbose_name, is_feminine

# Read the Django configuration
FORM_DEFAULT_TEMPLATE = getattr(settings, "FORM_DEFAULT_TEMPLATE")


DEFAULT_LIST_HTML = "<ul><li>&nbsp;</li></ul>"
DEFAULT_FORM_SECTION = "main"


class Form(dforms.Form):
    @classmethod
    def handle_request(cls, request: HttpRequest, **kwargs) -> HttpResponse:
        if request.method == 'POST':
            form = cls(data=request.POST, files=request.FILES, **kwargs)

            if not form.is_valid():
                form.display_errors(request)
            else: return form.submit(request)
        else: form = cls(**kwargs)

        return form.show(request)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self:
            self.transform_field(field)
        for field in self.get_text_list_fields():
            if field not in self.initial:
                self.initial.setdefault(field, DEFAULT_LIST_HTML)
            elif is_null_or_empty(self.initial[field]):
                self.initial[field] = DEFAULT_LIST_HTML

    def order_fields(self, field_order):
        super().order_fields(field_order)
        self.add_dynamic_fields()

    def add_dynamic_fields(self):
        pass

    def transform_field(self, field: BoundField):
        field.section = DEFAULT_FORM_SECTION
        for sname, sfields in self.get_sections().items():
            if field.name in sfields:
                field.section = sname
                break

        field.visible = self.is_visible_by_default(field)

        if field.widget_type == "checkbox":
            field.display_label = False
            field.display_icon = False
        elif field.widget_type == "ckeditor" or field.widget_type == "ckeditoruploading":
            field.display_label = True
            field.display_icon = False
        else:
            field.display_label = True
            field.display_icon = True

        field.icon = self.get_icon_for_field(field)

        widget = field.field.widget

        if isinstance(widget, CheckboxInput):
            widget.attrs["class"] = "form-check-input"
        else:
            if isinstance(widget, TextInput):
                widget.attrs["autocomplete"] = "off"

            if isinstance(widget, Select):
                widget.attrs["class"] = "form-select"
            else:
                widget.attrs["class"] = "form-control"

            if isinstance(widget, Select2Mixin):
                # noinspection PyUnresolvedReferences
                widget.attrs["data-placeholder"] = self.get_field_placeholder(field)
            else: widget.attrs["placeholder"] = self.get_field_placeholder(field)

    # noinspection PyUnresolvedReferences
    def full_clean(self):
        super().full_clean()
        for field in self.get_text_list_fields():
            if get_oneline_string(self.cleaned_data[field]) == DEFAULT_LIST_HTML:
                self.cleaned_data[field] = ""
                if hasattr(self.instance, field):
                    setattr(self.instance, field, "")

    def get_title(self) -> str:
        return "Formulaire"

    def get_page_title(self) -> str:
        return self.get_title()

    def get_header(self) -> Optional[str or SafeText]:
        return None

    def get_submit_text(self) -> str:
        return "Envoyer"

    def get_sections(self) -> Dict[str, List[str]]:
        return {}

    def get_section(self, section: str) -> List[BoundField]:
        fields = []
        for field in self:
            if field.section == section:
                fields.append(field)
        return fields

    def is_visible_by_default(self, field: BoundField) -> bool:
        return True

    def get_icon_for_field(self, field: BoundField) -> str:
        return "fas fa-arrow-right"

    def get_field_placeholder(self, field: BoundField) -> str:
        return field.label + " ..."

    def get_text_list_fields(self) -> List[str]:
        return []

    def get_context(self) -> dict:
        return {
            "form": self
        }

    def show(self, request: HttpRequest) -> HttpResponse:
        return render(request, FORM_DEFAULT_TEMPLATE, context=self.get_context())

    def submit(self, request: HttpRequest) -> HttpResponse:
        return HttpResponseRedirect("/")

    def display_errors(self, request: HttpRequest):
        errors: ErrorDict = self.errors
        for key, value in errors.items():
            field_name = self[key].label
            elist: ErrorList = value
            elist_txt = "<br />".join([str(e) for e in elist])
            messages.error(request, SafeText(f"<strong>{field_name}</strong><br />{elist_txt}"))


class ModelForm(Form, dforms.ModelForm):
    @classmethod
    def fetch_instance_from_id(cls) -> bool:
        return True

    # noinspection PyShadowingBuiltins
    @classmethod
    def handle_request(cls, request: HttpRequest, id=None, **kwargs) -> HttpResponse:
        if id is not None:
            # noinspection PyUnresolvedReferences
            model_type = cls.Meta.model
            try:
                kwargs["instance"] = model_type.objects.get(pk=id)
            except model_type.DoesNotExist:
                model_name = get_verbose_name(model_type)
                raise Http404(f"Unable to find {model_name} instance")

        return super().handle_request(request, **kwargs)

    def __init__(self, *args, instance=None, **kwargs):
        super().__init__(*args, instance=instance, **kwargs)
        self.edit_form = instance is not None and instance.pk is not None

    def is_edit_form(self) -> bool:
        return self.edit_form

    def get_title(self) -> str:
        if self.is_edit_form():
            return f"Modification de {self.instance}"
        else:
            model_name = get_verbose_name(self._meta.model)
            pronoun = "une" if is_feminine(self._meta.model) else "un"
            return f"Création d'{pronoun} {model_name}"

    def get_submit_text(self) -> str:
        verb = "Modifier" if self.is_edit_form() else "Créer"
        pronoun = "la" if is_feminine(self._meta.model) else "le"
        model_name = get_verbose_name(self._meta.model)
        return f"{verb} {pronoun} {model_name}"

    def save(self, commit=True):
        result = super().save(commit=False)
        self.set_result_fields(result)
        if commit:
            result.save()
            self.save_m2m()
        return result

    # This can be overriden to statically set some resulting fields
    def set_result_fields(self, result):
        pass

    def submit(self, request: HttpRequest) -> HttpResponse:
        self.save()
        return super().submit(request)

    def send_confirmation_message(self, request: HttpRequest):
        model_name = get_verbose_name(self._meta.model).capitalize()
        verb_suffix = "e" if is_feminine(self._meta.model) else ""
        if self.is_edit_form():
            messages.info(request, f"{model_name} modifié avec succès")
        else:
            messages.success(request, f"{model_name} créé{verb_suffix} avec succès")

    def get_delete_url(self) -> Optional[str]:
        return None


class DeleteModelForm(Form):
    @classmethod
    def get_model_type(cls) -> Type[Model]:
        raise NotImplemented()

    # noinspection PyMethodOverriding,PyShadowingBuiltins,PyUnresolvedReferences
    @classmethod
    def handle_request(cls, request: HttpRequest, id: int, **kwargs) -> HttpResponse:
        model_type = cls.get_model_type()
        try:
            instance = model_type.objects.get(pk=id)
            return super().handle_request(request, instance=instance, **kwargs)
        except model_type.DoesNotExist:
            raise Http404(f"Unable to find {model_type.__name__} instance with ID {id}")

    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance

    def get_title(self) -> str:
        model_type = self.get_model_type()
        model_name = get_verbose_name(model_type)
        pronoun = "une" if is_feminine(model_type) else "un"
        return f"Suppression d'{pronoun} {model_name}"

    def get_header(self) -> Optional[str or SafeText]:
        return SafeText(f"Souhaitez vous vraiment supprimer {self.instance} ? <strong>Cette action est irréversible</strong>.")

    def get_submit_text(self) -> str:
        return "Confirmer la suppression"

    def submit(self, request: HttpRequest) -> HttpResponse:
        # Delete the instance when the form is submitted
        self.instance.delete()
        return super().submit(request)

    def send_confirmation_message(self, request: HttpRequest):
        model_type = self.get_model_type()
        model_name = get_verbose_name(model_type).capitalize()
        verb_suffix = "e" if is_feminine(model_type) else ""
        messages.info(request, f"{model_name} supprimé{verb_suffix} avec succès")



from typing import List, Optional

from django.core.mail import EmailMultiAlternatives
from django.http import QueryDict
from django.template import loader


def send_mail(recipients: List[str], subject: str, body: str, cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None):
    email_message = EmailMultiAlternatives(subject, body, None, recipients, cc=cc, bcc=bcc)
    email_message.attach_alternative(body, "text/html")
    email_message.send()


def send_mail_template(recipients: List[str], subject: str, body_template: str, context, cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None):
    body = loader.render_to_string(body_template, context)
    send_mail(recipients, subject, body, cc, bcc)


def get_mailto_url(recipients: List[str], subject: str, body: str, cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None) -> str:
    query = QueryDict(mutable=True)
    query["subject"] = subject
    query["body"] = body
    if cc is not None:
        query.setlist("cc", cc)
    if bcc is not None:
        query.setlist("bcc", bcc)
    return f"mailto:{','.join(recipients)}?{query.urlencode()}"


def get_mailto_url_template(recipients: List[str], subject: str, body_template: str, context, cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None) -> str:
    body = loader.render_to_string(body_template, context)
    return get_mailto_url(recipients, subject, body, cc, bcc)



import math
from typing import TypeVar, List

from django.db.models import QuerySet
from django.http import HttpRequest, QueryDict
from pursuitlib.parsing import parseint


DEFAULT_PAGE_SIZE = 20
DEFAULT_MAX_PREV = 2
DEFAULT_MAX_NEXT = 2

T = TypeVar("T")


def page_query(request: HttpRequest, query: QuerySet[T],
                     page_var: str = "page", page_size: int = DEFAULT_PAGE_SIZE,
                     max_prev: int = DEFAULT_MAX_PREV, max_next: int = DEFAULT_MAX_NEXT) -> (List[T], dict):
    page = parseint(request.GET.get(page_var), 1)
    total = query.count()
    last_page = math.ceil(total / page_size)
    page = max(1, min(page, last_page))

    data = query[(page - 1) * page_size:page * page_size].all()

    prev_count = min(page - 1, max_prev)
    prev_pages = range(page - prev_count, page)

    next_count = min(last_page - page, max_next)
    next_pages = range(page + 1, page + 1 + next_count)

    url_query = QueryDict(mutable=True)
    for key in request.GET.keys():
        if key == page_var:
            continue
        url_query.setlist(key, request.GET.getlist(key))
    url_query_prefix = f"?{url_query.urlencode()}&" if len(url_query) > 0 else "?"

    return data, {
        "prev": prev_pages,
        "current": page,
        "next": next_pages,
        "last": last_page,
        "url_query_prefix": url_query_prefix,
    }



from typing import Type

from django import forms
import django_select2.forms as select2

from django.db.models import Model
from django.utils.translation import get_language


# Generic

class DateInput(forms.TextInput):
    input_type = "date"


# Select2

def set_attrs(attrs):
    attrs['data-token-separators'] = '[",", ";"]'
    attrs['data-theme'] = 'bootstrap-5'
    attrs['data-language'] = get_language()


def hide_search(attrs):
    attrs.setdefault("data-minimum-results-for-search", "Infinity")


class SelectInputWidget(select2.Select2Widget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        set_attrs(attrs)
        hide_search(attrs)
        return attrs


class ModelSelectInputWidget(select2.ModelSelect2Widget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        set_attrs(attrs)
        hide_search(attrs)
        return attrs


class ListInputWidget(select2.Select2MultipleWidget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        set_attrs(attrs)
        return attrs


class ModelListInputWidget(select2.ModelSelect2MultipleWidget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        set_attrs(attrs)
        return attrs


class TagInputWidget(select2.ModelSelect2TagWidget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        set_attrs(attrs)
        return attrs

    def value_from_datadict(self, data, files, name):
        vtype = self.get_value_type()
        values = super().value_from_datadict(data, files, name)

        for i in range(len(values)):
            if not values[i].isdigit():  # If this is a newly created object
                if self.use_lowercase_names():
                    values[i] = values[i].lower()
                # noinspection PyUnresolvedReferences
                try:
                    values[i] = self.queryset.get(name__iexact=values[i]).pk  # If the object already exists, don't create a new one
                except vtype.DoesNotExist:
                    values[i] = self.queryset.create(name=values[i]).pk

        return values

    def get_value_type(self) -> Type[Model]:
        raise NotImplemented()

    def use_lowercase_names(self) -> bool:
        return True


class SuggestTagInputWidget(TagInputWidget):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        attrs["data-minimum-input-length"] = 0
        return attrs



from django.http import HttpRequest, HttpResponse
from django.shortcuts import render

from pursuitlib_django.api.utils import response_message


# The site-specific error templates are supposed to be present in "error/<error_code>.html"


def handle_error(request: HttpRequest, code: int, name: str) -> HttpResponse:
    # If the request accepts a web page, return a standard human-readable error page
    if request.accepts("text/html"):
        return render(request, f"error/{code}.html", status=code)

    # Default to responding with a JSON error
    return response_message(name, code)


def error_400(request: HttpRequest, exception: Exception) -> HttpResponse:
    return handle_error(request, 400, "Bad request")


def error_403(request: HttpRequest, exception: Exception) -> HttpResponse:
    return handle_error(request, 403, "Forbidden")


def error_404(request: HttpRequest, exception: Exception) -> HttpResponse:
    return handle_error(request, 404, "Not found")


def error_500(request: HttpRequest) -> HttpResponse:
    return handle_error(request, 500, "Internal server error")



import os
from typing import List


def delete_media(model, fields: List[str]):
    for field in fields:
        value = getattr(model, field)

        if value:
            if os.path.isfile(value.path):
                os.remove(value.path)


def update_media(model, fields: List[str]):
    if not model.pk:
        return False

    try:
        old_model = type(model).objects.get(pk=model.pk)

        for field in fields:
            value = getattr(model, field)
            old_value = getattr(old_model, field)

            if old_value != value:
                if old_value and os.path.isfile(old_value.path):
                    os.remove(old_value.path)
    except type(model).DoesNotExist:
        return False



import os
from typing import Optional
from typing import Type

from django.conf import settings
from django.db import ProgrammingError
from django.db.models import Model, FileField

# Read the Django configuration
MEDIA_ROOT = getattr(settings, "MEDIA_ROOT")


def get_media_path(media: FileField) -> Optional[str]:
    return os.path.join(MEDIA_ROOT, media.name) if media else None


def get_model_ctype_name(model: Type[Model]) -> str:
    return model.__name__.lower()


# noinspection PyProtectedMember
def get_verbose_name(obj) -> str:
    try:
        return obj._meta.verbose_name
    except AttributeError:
        return type(obj).__name__


# noinspection PyProtectedMember
def get_verbose_name_plural(obj) -> str:
    try:
        return obj._meta.verbose_name_plural
    except AttributeError:
        return get_verbose_name(obj) + "s"


def is_feminine(obj) -> bool:
    try:
        # noinspection PyTypeChecker
        if not isinstance(obj, Type):
            obj = type(obj)
        return obj.Gender.feminine
    except AttributeError:
        return False


def safemap(function, input_set) -> map:  # Used to create maps without throwing an error when the database has not been yet initialized
    try:
        return map(function, input_set)
    except ProgrammingError:  # If the database has not been yet initialized, return an empty list, this souldn't matter anyway
        return map(lambda x: x, [])



from django.http import JsonResponse


# Generic responses

def response_success(message: str = "The operation was completed successfully", status: int = 200) -> JsonResponse:
    return response_message(message, status)


def response_message(message: str, status: int = 200) -> JsonResponse:
    return response_object({"message": message}, status)


def response_object(obj: dict, status: int = 200) -> JsonResponse:
    return JsonResponse(obj, status=status)



from typing import Type

from django.contrib.admin import SimpleListFilter
from django.db.models import Model, ForeignKey


def foreign_key_filter(model: Type[Model], field_name: str):
    field: ForeignKey = getattr(model, field_name).field

    class ForeignKeyFilter(SimpleListFilter):
        title = field.verbose_name.capitalize()
        parameter_name = field_name

        def lookups(self, request, model_admin):
            # noinspection PyUnresolvedReferences
            return [(instance.pk, str(instance)) for instance in field.related_model.objects.all()]

        def queryset(self, request, queryset):
            if not self.value():
                return queryset

            query = {
                f"{field_name}__pk": self.value()
            }

            return queryset.filter(**query)

    return ForeignKeyFilter




from hashlib import md5


def get_avatar_url(email: str, default="mp") -> str:
    email_hash = md5(email.strip().lower().encode("utf-8")).hexdigest()
    return f"https://www.gravatar.com/avatar/{email_hash}?d={default}"



from typing import Type

from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied, BadRequest
from django.http import HttpResponse, HttpRequest
from pursuitlib.decorators import decorator
from pursuitlib.parsing import parseas

from pursuitlib_django.forms import Form


# View decorators

def staff_required(function):
    def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        # noinspection PyUnresolvedReferences
        if request.user.is_staff:
            return function(request, *args, **kwargs)
        else:
            raise PermissionDenied("Only the staff can access this resource")
    return wrapper


def get_form_decorator(base_decorator):
    def form_decorator(function, *dargs, **dkwargs):
        # request needs to be the first argument for compatibility with the view decorator
        def remap_wrapper(request: HttpRequest, form_type: Type[Form], *args, **kwargs) -> HttpResponse:
            return function(form_type, request, *args, **kwargs)

        result = base_decorator(remap_wrapper, *dargs, **dkwargs)

        # Remapping input arguments
        return lambda form_type, request, *args, **kwargs: result(request, form_type, *args, **kwargs)
    return form_decorator


form_login_required = get_form_decorator(login_required)
form_staff_required = get_form_decorator(staff_required)


# Action decorators

@decorator
def action(function, *params, method: str = "POST"):
    def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if request.method != method:
            raise BadRequest(f"Only {method} requests are allowed here")

        if method == "POST":
            request_data = request.POST
        elif method == "GET":
            request_data = request.GET
        else:
            raise Exception(f"Unsupported HTTP method: {method}")

        data = {}
        for param in params:
            # Cast the parameter in the appropriate type
            if ':' in param:
                parts = param.split(':')
                type_name = parts[0]
                name = parts[1]
            else:
                type_name = "str"
                name = param

            if name not in request_data:
                raise BadRequest(f"Missing '{name}' {method} parameter")
            data[name] = parseas(type_name, request_data[name])
            if data[name] is None:
                raise BadRequest(f"Invalid '{name}' {method} parameter")

        kwargs |= data
        return function(request, *args, **kwargs)

    return wrapper




from typing import TypeVar, Type, Optional, Union

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models.fields import Field
from django.db.models.fields.files import FileField, ImageField
from django.forms import forms
from pursuitlib.indexedenum import IndexedEnum
from pursuitlib.utils import is_null_or_empty

# Read the Django configuration
MAX_UPLOAD_SIZE = getattr(settings, "MAX_UPLOAD_SIZE", 5)  # The default max upload size is 5 MiB


MAX_UPLOAD_SIZE_BYTES = MAX_UPLOAD_SIZE * 1024 * 1024  # Convert from MiB to bytes
UPLOAD_ERROR_MSG = f"Le fichier téléversé est trop grand. Il doit avoir une taille inférieure à {MAX_UPLOAD_SIZE} Mo"


T = TypeVar("T", bound=IndexedEnum)


class EnumField(Field):
    @staticmethod
    def get_choices_for(enum: Type[T]):
        return [(item, str(item.value)) for item in enum]

    def __init__(self, enum: Type[T], **kwargs):
        super().__init__(choices=EnumField.get_choices_for(enum), **kwargs)
        self.enum = enum

    def get_internal_type(self):
        return "IntegerField"

    # Create a DB value from an enum.
    # Integers are also accepted as they're easier to use in migration files
    def get_prep_value(self, value: Union[T, int]):
        if is_null_or_empty(value):
            return None
        if isinstance(value, int):
            value = self.get_at(value)
        return super().get_prep_value(value.ordinal)

    # Create an enum from a DB value
    def from_db_value(self, value, expression, connection) -> Optional[T]:
        if value is None:
            return None
        return self.get_at(int(value))

    # Try to get an enum from another value
    def to_python(self, value) -> Optional[T]:
        if is_null_or_empty(value):
            return None
        if isinstance(value, self.enum):
            return value

        try:
            return self.enum[value]
        except KeyError:
            raise ValidationError("Invalid enum value")

    def deconstruct(self):
        name, path, lst, keywords = super().deconstruct()

        # Choices are handled by the "enum" parameter
        del keywords["choices"]
        keywords["enum"] = self.enum

        return name, path, lst, keywords

    def get_at(self, index: int) -> T:
        # Clamp the index to avoid raising errors
        length = len(self.enum)
        index = max(0, min(index, length - 1))
        return self.enum.values()[index]


class FileUploadField(FileField):
    def clean(self, *args, **kwargs):
        data = super().clean(*args, **kwargs)
        if data.file.size > MAX_UPLOAD_SIZE_BYTES:
            raise forms.ValidationError(UPLOAD_ERROR_MSG)
        return data


class ImageUploadField(ImageField):
    def clean(self, *args, **kwargs):
        data = super().clean(*args, **kwargs)
        if data.file.size > MAX_UPLOAD_SIZE_BYTES:
            raise forms.ValidationError(UPLOAD_ERROR_MSG)
        return data



import json

from django.core.exceptions import BadRequest, PermissionDenied
from django.http import HttpRequest, HttpResponse
from pursuitlib.decorators import decorator


@decorator
def api_get(function, public = False):
    def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if request.method != "GET":
            raise BadRequest(f"Only GET requests are allowed")
        # noinspection PyUnresolvedReferences
        if not public and request.user.is_anonymous:
            raise PermissionDenied()
        return function(request, *args, **kwargs)
    return wrapper


@decorator
def api_post(function, public = False):
    def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if request.method != "POST":
            raise BadRequest(f"Only POST requests are allowed")
        # noinspection PyUnresolvedReferences
        if not public and request.user.is_anonymous:
            raise PermissionDenied()

        try:
            data = json.loads(request.body)
        except:
            raise BadRequest("Invalid JSON payload")

        return function(request, *args, data=data, **kwargs)
    wrapper.csrf_exempt = True
    return wrapper
